/*                                                                                                                                
   minexample.cu
               
   - assigns some numbers to an image downloads the result
     and saves it
                                                                                                                                  
   (c) 2012 Ivo Ihrke                                                                                                        
                                                                                                                                  
   Universitaet des Saarlandes / MPI Informatik                                                                                   
                                                                                                                                  
   ihrke@mmci.uni-saarland.de                                                                                                     
   
   
   thanks to an initial version 

   Copyright (C) 2010 Bastian Goldluecke,
                                                                                                                               
*/                                                                                                                                

#define MAIN_PROG

#include <iostream>
#include <stdio.h>
#include <assert.h>

using namespace std;


// Our image class
#include "../../Image/pvc_image.h"


#define cuda_float float

// constants for the width and height of our image
#define W 256
#define H 256

// CUDA Block size
#define BLOCK_SIZE_X 16
#define BLOCK_SIZE_Y 16

// Macro to handle CUDA error return
#define CUDA_SAFE_CALL(c) if ( c != cudaSuccess ) { printf("CUDA error: %s, line %d\n", cudaGetErrorString( cudaGetLastError() ), __LINE__); assert( false ); }




// example kernel
__global__ void example_kernel( cuda_float *px ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  px [ off + 0 ] = (cuda_float)ox / (cuda_float)W;
  px [ off + 1 ] = (cuda_float)oy / (cuda_float)H;
  px [ off + 2 ] = 0.5;
  px [ off + 3 ] = 1.0;

     
  __syncthreads();
       

}


// this one checks if your card is suitable and drivers are working
int init_CUDA()
{
  // Check available devices
  int error = 0;
  int deviceCount;
  int gpuDeviceCount = 0;
  struct cudaDeviceProp deviceProp;
  
  cudaError_t result = cudaGetDeviceCount( &deviceCount );

  if ( result != cudaSuccess )
    {
      cout << "No CUDA devices found - Error string: " << cudaGetErrorString( result ) << endl;
      
      // machines without GPU should still support emulation
      deviceCount = 0; 
      
      error = -1;
    }
  
  for ( int device = 0; device < deviceCount; ++device ) 
    {
      cudaGetDeviceProperties(&deviceProp, device);
      
      if ( deviceProp.major != 9999) /* 9999 means emulation only */
	++gpuDeviceCount;
    }
  
  cout <<  gpuDeviceCount << " GPU CUDA device(s) found" << endl;

  // Use first device found
  cudaSetDevice( 0 );
  cudaGetDeviceProperties( &deviceProp, 0 );
  cout << "Using CUDA device " << deviceProp.name << endl;

  return error;
}








  // --------------------------------------------------------------------------------
  // ------------------------------------- main function ----------------------------
  // --------------------------------------------------------------------------------

// Main function for example
int main( int argc, char *argv [] )
{
  
  
  // this is where our result will be stored
  PVC_Image  *result = NULL;

  // a CUDA GPU-based array 
  cuda_float *gpu_px = NULL;

  // report errors in this variable
  cudaError_t err;

  // number of pixels
  int N = W * H;

  if ( init_CUDA() < 0 )
    {
      fprintf( stderr,"Could not initialize CUDA, exiting ...\n" );
      exit ( -1 );
    }


  // allocate N * 4 floats on the GPU (4 color channels)
  CUDA_SAFE_CALL( cudaMalloc( &gpu_px, N * sizeof( cuda_float ) * 4 ) );

  // initialize our result image
  result = new PVC_Image();
  result -> resize ( W, H );

  // Set block size (in number of threads)
  dim3 dimBlock( BLOCK_SIZE_X, BLOCK_SIZE_Y );

  // Set grid size (in number of blocks)
  dim3 dimGrid( W / dimBlock.x, H / dimBlock.y ); 

  // Perform computation
  cout << "GPU computation: ["; cout.flush();
  clock_t gpu_start = clock();


  
  // issue the computation
  example_kernel <<< dimGrid, dimBlock >>> ( gpu_px );

  // make sure all threads are done
  cudaThreadSynchronize();

  // did it work ?
  err = cudaGetLastError();
  if (err != cudaSuccess) 
    printf("Error: %s\n", cudaGetErrorString(err));  

  // timing
  clock_t gpu_end = clock();

  cout << "] done." << endl;
  cout << "GPU time: " << (gpu_end - gpu_start) / double(CLOCKS_PER_SEC) << endl;  
  

  // Copy back result to CPU
  // we can read directly to our image class because 
  // we have non-padded memory (just W*H floats)
  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_px , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );

  // Save results

  result -> save( "Output/minexample.ppm" );

  // Cleanup
  // GPU
  CUDA_SAFE_CALL( cudaFree( gpu_px ));

  // Host
  delete result;

  return 0;
}

