/*                                                                                                                                
   oflow.cu
               
   - implements Horn & Schunck optical flow
                                                                                                                                  
   (c) 2012 Ivo Ihrke                                                                                                        
                                                                                                                                  
   Universitaet des Saarlandes / MPI Informatik                                                                                   
                                                                                                                                  
   ihrke@mmci.uni-saarland.de                                                                                                     
                                                                                                                                  
*/                                                                                                                                

#define MAIN_PROG

#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <assert.h>
//#include <unistd.h>

#include <vector>

using namespace std;


// Our image class
#include "../../Image/pvc_image.h"


// this one makes it easier to exchange float for double 
// on devices of higher compute capability if more accuracy 
// is needed 
#define cuda_float float

int W;
int H;

// CUDA Block size
#define BLOCK_SIZE_X 16
#define BLOCK_SIZE_Y 16

// Macro to handle CUDA error return
#define CUDA_SAFE_CALL(c) if ( c != cudaSuccess ) { cout << "CUDA error: " << cudaGetErrorString( cudaGetLastError() )  << ", line " << __LINE__ << endl; cout.flush(); assert( false ); }


// --------------------------------------------------------------------------------
// ---------------------------------------- CUDA kernels  -------------------------
// --------------------------------------------------------------------------------


// sets all values in <array> to <value>
__global__ void set_value( cuda_float *array, int W, int H, cuda_float value ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ox + oy * W;

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( off >= 0 &&  off < W * H )
    array [ off ] = value;

}




// example kernel that reads input
__global__ void gauss_sep_hor( cuda_float *px, cuda_float *out, const int W, const int H, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H * 4 )
    return;
  
  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int i = -halfsize; i <= halfsize; i++ )
	  { 
	    int img_off = i * 4;

	    // use center value if kernel is outside image
	    if ( ox + i < 0 || ox + i > W - 1 )
	      img_off = 0;

	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = i * i;
	    
	    cuda_float kval = exp( - square * SIGMA_square );

	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }
  
      
}


// example kernel that reads input
__global__ void gauss_sep_ver( cuda_float *px, cuda_float *out, const int W, const int H, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H * 4 )
    return;

  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int j = -halfsize; j <= halfsize; j++ )
	  { 
	    int img_off = j * W * 4;

	    if ( oy + j < 0 || oy + j > H - 1 )
	      img_off = 0;
	    
	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = j * j;
	    
	    //	    cuda_float kval = exp( - square / ( 2.0 * SIGMA * SIGMA ) );
	    cuda_float kval = exp( - square * SIGMA_square );


	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }

}


//helper functions -- a device function is always inlined for architectures below 3.5 or so

//returns luminance of a color image
__device__ cuda_float get_luminance( cuda_float *colimg, int imgoff, int W, int H ) 
{
  return 0.299f * colimg [ imgoff + PVC_CHANNEL_R ] + 0.587f * colimg [ imgoff + PVC_CHANNEL_G ] + 0.114f * colimg [ imgoff + PVC_CHANNEL_B ];
}

//returns "average" of a scalar field (single channel image) - required in classical formulation of Horn-Schunck
__device__ cuda_float average_classic( cuda_float *img, int off, int W, int H ) 
{

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return 0.0;

  return 1.0/12.0 * img [ off - 1 - W ] +  1.0/6.0 * img[ off - W ] + 1.0/12.0 * img [ off + 1 - W ] +
         1.0/6.0  * img [ off - 1     ] +                             1.0/6.0  * img [ off + 1     ] +
         1.0/12.0 * img [ off - 1 + W ] +  1.0/6.0 * img[ off + W ] + 1.0/12.0 * img [ off + 1 + W ];
}


// the optical flow kernel
__global__ void horn_schunck_classic_IxIyIt( cuda_float *left, cuda_float *right, int W, int H, cuda_float *Ix, cuda_float *Iy, cuda_float *It ) 
{

  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );
  int imgoff = off * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( ox >= 1 && ox < W - 1 )
    {
      // over deltaX  - choose deltaX and deltaY equal, ignore overall scale factor
      Ix [ off ] = 1.0f / 2.0f * ( get_luminance(  left, imgoff + 4, W, H ) - 
				   get_luminance(  left, imgoff - 4, W, H ) +
				   get_luminance( right, imgoff + 4, W, H ) - 
				   get_luminance( right, imgoff - 4, W, H )  );

    }
  else
    Ix [ off ] = 0.0f; // von Neumann = 0
  

  if ( oy >= 1 && oy < H - 1 )
    {
      // over deltaX  - choose deltaX and deltaY equal, ignore overall scale factor
      Iy [ off ] = 1.0f / 2.0f * ( get_luminance(  left, imgoff + 4 * W, W, H ) - 
				   get_luminance(  left, imgoff - 4 * W, W, H ) + 
				   get_luminance( right, imgoff + 4 * W, W, H ) -
				   get_luminance( right, imgoff - 4 * W, W, H ) ); 
    }
  else
    Iy [ off ] = 0.0f; // von Neumann = 0

  // ( I_{t+1} - I_{t} ) / deltaT -> again, ignore deltaT
  It [ off ] = ( get_luminance( right, imgoff, W, H ) -  get_luminance( left, imgoff, W, H ) );

}

__global__ void horn_schunck_classic_ubarvbar( cuda_float *u, cuda_float *v, int W, int H, cuda_float *ubar, cuda_float *vbar ) 
{
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( ox > 0 && ox < W - 1 &&
       oy > 0 && oy < H - 1 )
    {
      ubar[ off ] = average_classic( u, off, W, H );
      vbar[ off ] = average_classic( v, off, W, H );
    }
  else
    {
      //default - corners
      ubar [ off ] = 0.0;
      vbar [ off ] = 0.0;

      // left
      if ( ox == 0 && oy != 0 && oy != H - 1 )
	{
	  ubar[ off ] = average_classic( u, off + 1, W, H );
	  vbar[ off ] = average_classic( v, off + 1, W, H );
	}

      // right
      if ( ox == W - 1 && oy != 0 && oy != H - 1 )
	{
	  ubar[ off ] = average_classic( u, off - 1, W, H );
	  vbar[ off ] = average_classic( v, off - 1, W, H );
	}


      // top
      if ( oy == 0 && ox != 0 && ox != W - 1 )
	{
	  ubar[ off ] = average_classic( u, off + W, W, H );
	  vbar[ off ] = average_classic( v, off + W, W, H );
	}

      // bottom
      if ( oy == H - 1 && ox != 0 && ox != W - 1 )
	{
	  ubar[ off ] = average_classic( u, off - W, W, H );
	  vbar[ off ] = average_classic( v, off - W, W, H );
	}

    }
}

__global__ void horn_schunck_classic_take_step( cuda_float *u, cuda_float *v, int W, int H, cuda_float *ubar, cuda_float *vbar, cuda_float *Ix, cuda_float *Iy, cuda_float *It, float ALPHA ) 
{

  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;
  
  int off    = ( ox + oy * W );
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  cuda_float vx  = Ix [ off ];
  cuda_float vy  = Iy [ off ];

  cuda_float val = ( vx * ubar[ off ]  +
		     vy * vbar[ off ]  +
		     It [ off ]  ) / ( ALPHA * ALPHA + vx * vx + vy * vy );

  u [ off ]  = ubar [ off ] - vx * val;
  
  v [ off ]  = vbar [ off ] - vy * val;
  
}




// the visualization kernel - takes the scalar fields u and v (components of optical flow)
// and turns it into a color image in the range [0,1]
__global__ void visualize( cuda_float *u, cuda_float *v, cuda_float *visresult, int W, int H )
{
  cuda_float MAXLENGTH = 5.0;
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );
  int imgoff = off * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  // simple 
  visresult [ imgoff + PVC_CHANNEL_R ] = u [ off ] / MAXLENGTH + 0.5;
  visresult [ imgoff + PVC_CHANNEL_G ] = v [ off ] / MAXLENGTH + 0.5;
  visresult [ imgoff + PVC_CHANNEL_B ] = 0.0;
  visresult [ imgoff + PVC_CHANNEL_A ] = 1.0;

  // HSV encoding
  if ( 0 )
    {
      cuda_float length = sqrt( u [ off ] * u [ off ] + v [ off ] * v [ off ] );
      
      cuda_float cosa = u [ off ] / length;
      cuda_float sina = v [ off ] / length;
      
      cuda_float angle = atan2( cosa, sina );   //C
      
      cuda_float value = length / MAXLENGTH;    //H 
      
      cuda_float angle_dash = angle / (3.14159f / 3.0f );
      
      if ( 0.0f <= angle_dash && angle_dash < 1.0f )
	{
	  visresult [ imgoff + 2 ] = u [ off ] * 10.0 + 0.5;
	  visresult [ imgoff + 1 ] = v [ off ] * 10.0 + 0.5;
	  visresult [ imgoff + 0 ] = 0.0;
	  visresult [ imgoff + 3 ] = 1.0;
	  
	}
    }
} 




//returns bilinearly interpolated pixel value
__device__ cuda_float get_pixel_bilinear( cuda_float *img, cuda_float x, cuda_float y, int W, int H, int c ) 
{
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return 0.0;

  if ( x < 0 || x > W - 1 || y < 0 || y > H - 1 )
    return 0.0;

  off = ( floor( x ) + floor( y ) * W ) * 4 + c;

  cuda_float Iul, Iur, Ill, Ilr;

  Iul = img [ off ];
  Iur = img [ off + 4 ];
  Ill = img [ off + 4 * W ];
  Ilr = img [ off + 4 * W + 4 ];

  cuda_float ax, ay;
  
  ax =  x - floor( x );
  ay =  y - floor( y );

  cuda_float Iux, Ilx;

  Iux = Iul * ( 1.0f - ax ) + Iur * ax;
  Ilx = Ill * ( 1.0f - ax ) + Ilr * ax;
  
  return ( 1.0f - ay ) * Iux + ay * Ilx;
}





// the warp kernel - uses the scalar fields u and v (components of optical flow)
// to warp the from image 
__global__ void warp( cuda_float *u, cuda_float *v, cuda_float *from, cuda_float *visresult, int W, int H )
{
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  cuda_float uval = ox + u [ off ] * 2;
  cuda_float vval = oy + v [ off ] * 2;


  int imgoff       = off * 4;

  if ( uval > 0 && uval < W - 1 && 
       vval > 0 && vval < H - 1 )
    {
      /*
	int imglookupoff = ( ox + uval + (oy +vval) * W ) * 4; 
	
	// nearest
	visresult [ imgoff + PVC_CHANNEL_R ] = from  [ imglookupoff + PVC_CHANNEL_R ];
	visresult [ imgoff + PVC_CHANNEL_G ] = from  [ imglookupoff + PVC_CHANNEL_G ];
	visresult [ imgoff + PVC_CHANNEL_B ] = from  [ imglookupoff + PVC_CHANNEL_B ];
	visresult [ imgoff + PVC_CHANNEL_A ] = from  [ imglookupoff + PVC_CHANNEL_A ];
      */

	// bilinear
      visresult [ imgoff + PVC_CHANNEL_R ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_R  );
      visresult [ imgoff + PVC_CHANNEL_G ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_G  );
      visresult [ imgoff + PVC_CHANNEL_B ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_B  );
      visresult [ imgoff + PVC_CHANNEL_A ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_A  );

    }
  else
    {
      visresult [ imgoff + PVC_CHANNEL_R ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_G ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_B ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_A ] = 0.0f;
    }
} 




// --------------------------------------------------------------------------------
// ------------------------------------- CUDA init stuff  -------------------------
// --------------------------------------------------------------------------------





// from CUDA SDK
// Beginning of GPU Architecture definitions
inline int _ConvertSMVer2Cores(int major, int minor)
{
  // Defines for GPU Architecture types (using the SM version to determine the # of cores per SM
  typedef struct {
    int SM; // 0xMm (hexidecimal notation), M = SM Major version, and m = SM minor version
    int Cores;
  } sSMtoCores;

  sSMtoCores nGpuArchCoresPerSM[] = 
    { { 0x10,  8 },
      { 0x11,  8 },
      { 0x12,  8 },
      { 0x13,  8 },
      { 0x20, 32 },
      { 0x21, 48 },
      {   -1, -1 } 
    };

  int index = 0;
  while (nGpuArchCoresPerSM[index].SM != -1) {
    if (nGpuArchCoresPerSM[index].SM == ((major << 4) + minor) ) {
      return nGpuArchCoresPerSM[index].Cores;
    }
    index++;
  }
  printf("MapSMtoCores undefined SMversion %d.%d!\n", major, minor);
  return -1;
}
// end of GPU Architecture definitions




// this one checks if your card is suitable and drivers are working
int init_CUDA()
{
  // Check available devices
  int error = 0;
  int deviceCount;
  int gpuDeviceCount = 0;
  struct cudaDeviceProp deviceProp;
  
  cudaError_t result = cudaGetDeviceCount( &deviceCount );

  if ( result != cudaSuccess )
    {
      cout << "No CUDA devices found - Error string: " << cudaGetErrorString( result ) << endl;
      
      // machines without GPU should still support emulation
      deviceCount = 0; 
      
      error = -1;
    }
  
  for ( int device = 0; device < deviceCount; ++device ) 
    {
      CUDA_SAFE_CALL ( cudaGetDeviceProperties(&deviceProp, device) );
      
      if ( deviceProp.major != 9999) /* 9999 means emulation only */
	++gpuDeviceCount;
    }
  
  cout <<  gpuDeviceCount << " GPU CUDA device(s) found" << endl;

  // Use first device found - CUDA only
  CUDA_SAFE_CALL ( cudaSetDevice( 0 ) );

  // Use first device found - CUDA and GL
  //  CUDA_SAFE_CALL ( cudaGLSetGLDevice( 0 ) );

  CUDA_SAFE_CALL ( cudaSetDeviceFlags(  cudaDeviceMapHost ) );

  cudaGetDeviceProperties( &deviceProp, 0 );
  cout << "Using CUDA device " << deviceProp.name << endl;

  printf("[%s] has %d MP(s) x %d (Cores/MP) = %d (Cores)\n",
	 deviceProp.name, deviceProp.multiProcessorCount,
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor),
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor) * deviceProp.multiProcessorCount);

  cout << "ClockSpeed: " << deviceProp.clockRate << " Hz (?)." << endl;

  cout.flush();

  return error;
}


int cudaCheckError( char *desc )
{
  // report errors in this variable
  cudaError_t err;
  
  if ( (err = cudaGetLastError() ) != cudaSuccess ) 
    {
      cout << "Error at " << desc << ": " <<  cudaGetErrorString( err ) << endl;  
      cout.flush();
    }
  return err;
  
}






// --------------------------------------------------------------------------------
// ------------------------------------- main function ----------------------------
// --------------------------------------------------------------------------------

// Main function for example
int main( int argc, char *argv [] )
{
  
  
  // this is where our result will be stored
  PVC_Image  *result       = NULL;
  PVC_Image  *input_left   = NULL;
  PVC_Image  *input_right  = NULL;

  // a CUDA GPU-based array 
  cuda_float *gpu_left   = NULL;
  cuda_float *gpu_right  = NULL;

  cuda_float *gpu_u      = NULL;
  cuda_float *gpu_v      = NULL;

  cuda_float *gpu_ubar   = NULL;
  cuda_float *gpu_vbar   = NULL;

  cuda_float *gpu_Ix     = NULL;
  cuda_float *gpu_Iy     = NULL;
  cuda_float *gpu_It     = NULL;


  cuda_float *gpu_dum    = NULL;
  cuda_float *gpu_out    = NULL;

  
  // ------------ *NEW* begin --------------
  input_left = new PVC_Image;
  //input_left -> load("../../Data/teddy_left.ppm");
  //input_left -> load("a.ppm");
  input_left -> load("../../Data/yosemite10.ppm");

  input_right = new PVC_Image;
  //input_right -> load("../../Data/teddy_right.ppm");
  //input_right -> load("b.ppm");
  input_right -> load("../../Data/yosemite09.ppm");

  // number of pixels
  W = input_left -> w();
  H = input_left -> h();
  int N = W * H;

  if ( input_left -> w() != input_right -> w() ||
       input_left -> h() != input_right -> h() )
    {
      cout << "Image size do not match, exiting ..." << endl;
      cout.flush();
      exit( -1 );
    }
       

  if ( init_CUDA() < 0 )
    {
      fprintf( stderr,"Could not initialize CUDA, exiting ...\n" );
      exit ( -1 );
    }




  // allocate N * 4 floats on the GPU (4 color channels) - will be the input now
  CUDA_SAFE_CALL( cudaMalloc( &gpu_left, N * sizeof( cuda_float ) * 4 ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_right, N * sizeof( cuda_float ) * 4 ) );

  // flow vectors 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_u, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_v, N * sizeof( cuda_float ) ) );

  // average flow 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_ubar, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_vbar, N * sizeof( cuda_float ) ) );

  // gradients
  CUDA_SAFE_CALL( cudaMalloc( &gpu_Ix, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_Iy, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_It, N * sizeof( cuda_float ) ) );


  // allocate output 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_dum, N * sizeof( cuda_float ) * 4 ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_out, N * sizeof( cuda_float ) * 4 ) );

  
  // initialize our result image
  result = new PVC_Image;
  result -> resize( W, H );


  // Set block size (in number of threads)
  dim3 dimBlock( BLOCK_SIZE_X, BLOCK_SIZE_Y );
  
  // Set grid size (in number of blocks)
  dim3 dimGrid( W / dimBlock.x, H / dimBlock.y ); 



  // initialize result arrays 
  set_value <<< dimGrid, dimBlock >>> ( gpu_u, W, H, 0.0f );
  cudaCheckError( "setvalue u" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_v, W, H, 0.0f );
  cudaCheckError( "setvalue v" );

  set_value <<< dimGrid, dimBlock >>> ( gpu_ubar, W, H, 0.0f );
  cudaCheckError( "setvalue ubar" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_vbar, W, H, 0.0f );
  cudaCheckError( "setvalue vbar" );

  set_value <<< dimGrid, dimBlock >>> ( gpu_Ix, W, H, 0.0f );
  cudaCheckError( "setvalue Ix" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_Iy, W, H, 0.0f );
  cudaCheckError( "setvalue Iy" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_It, W, H, 0.0f );
  cudaCheckError( "setvalue It" );


  

  // Perform computation
  cerr << "GPU computation: ["; cerr.flush();
  clock_t gpu_start = clock();

#define ALPHA 1.0f
#define INNER_LOOPS 2000
 
  float SIGMA = 0.8;
  int HS = (int) ceil( SIGMA ) * 2 + 1;

  cerr << "SIGMA: " << SIGMA << " HalfSize: " << HS << endl;

      
  // Copy matrices to GPU (host to device)
  CUDA_SAFE_CALL( cudaMemcpy( gpu_left,  input_left -> get_data(), N * sizeof(cuda_float) * 4, cudaMemcpyHostToDevice ));
  CUDA_SAFE_CALL( cudaMemcpy( gpu_right, input_right -> get_data(), N * sizeof(cuda_float) * 4, cudaMemcpyHostToDevice ));


  // pre-blur left and right images
  gauss_sep_hor <<< dimGrid, dimBlock >>> ( gpu_left, gpu_dum, W, H, HS, SIGMA );
  cudaCheckError( "horizontal preblur left" );
  cudaThreadSynchronize();
  gauss_sep_ver <<< dimGrid, dimBlock >>> ( gpu_dum, gpu_left, W, H, HS, SIGMA );
  cudaCheckError( "vertical preblur left" );
      
  gauss_sep_hor <<< dimGrid, dimBlock >>> ( gpu_right, gpu_dum, W, H, HS, SIGMA );
  cudaCheckError( "horizontal preblur right" );
  cudaThreadSynchronize();
  gauss_sep_ver <<< dimGrid, dimBlock >>> ( gpu_dum, gpu_right, W, H, HS, SIGMA );
  cudaCheckError( "vertical preblur right" );

  // compute derivatives
  horn_schunck_classic_IxIyIt <<< dimGrid, dimBlock >>> ( gpu_left, gpu_right, W, H, gpu_Ix, gpu_Iy, gpu_It );
  cudaCheckError( "compute gradients (before inner loop)" );

      
  // main iteration
  for ( int i = 0; i < INNER_LOOPS; i++ )
    {
      
      // compute averages
      horn_schunck_classic_ubarvbar <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, W, H, gpu_ubar, gpu_vbar );
      cudaCheckError( "Horn-Schunck ubar vbar" );
      
      // compute main step 
      horn_schunck_classic_take_step <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, W, H, gpu_ubar, gpu_vbar, gpu_Ix, gpu_Iy, gpu_It, ALPHA );
      cudaCheckError( "Horn-Schunck main step" );

    }


  // timing
  clock_t gpu_end = clock();

  cout << "] done." << endl;
  cout << "GPU time: " << (gpu_end - gpu_start) / double(CLOCKS_PER_SEC) << endl;  
  


  // Copy back result to CPU
  // we can read directly to our image class because 
  // we have non-padded memory (just W*H floats)

  // Save results
  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_left , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_left.ppm" );

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_right , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_right.ppm" );

  visualize <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, gpu_left, W, H );
  cudaCheckError( "visualize" );

  cudaThreadSynchronize();

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_left , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_flow.ppm" );

  // use (u,v) to warp gpu_right such that it matches gpu_left -> store result in gpu_left
  warp <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, gpu_right, gpu_out, W, H );
  cudaCheckError( "warp" );

  cudaThreadSynchronize();

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_out , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_warp.ppm" );


  // Cleanup
  // GPU
  CUDA_SAFE_CALL( cudaFree( gpu_left ));
  CUDA_SAFE_CALL( cudaFree( gpu_right ));
  CUDA_SAFE_CALL( cudaFree( gpu_u ));
  CUDA_SAFE_CALL( cudaFree( gpu_v ));
  CUDA_SAFE_CALL( cudaFree( gpu_ubar ));
  CUDA_SAFE_CALL( cudaFree( gpu_vbar ));
  CUDA_SAFE_CALL( cudaFree( gpu_Ix ));
  CUDA_SAFE_CALL( cudaFree( gpu_Iy ));
  CUDA_SAFE_CALL( cudaFree( gpu_It ));

  CUDA_SAFE_CALL( cudaFree( gpu_dum ));
  CUDA_SAFE_CALL( cudaFree( gpu_out ));

  // Host
  delete result;
  delete input_left;
  delete input_right;

  return 0;
}

