/*                                                                                                                                
   oflow_v3.cu
               
   - implements Horn & Schunck optical flow
   - with OpenGL visualization (using deprecated registration of PBOs)
           
   - use #define SHOW to toggle visualization
                                                                                                                       
   (c) 2012 Ivo Ihrke                                                                                                        
                                                                                                                                  
   Universitaet des Saarlandes / MPI Informatik                                                                                   
                                                                                                                                  
   ihrke@mmci.uni-saarland.de                                                                                                     
                                                                                                                                  
*/                                                                                                                                

#define MAIN_PROG

#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <assert.h>
//#include <unistd.h>

#include <vector>

using namespace std;


// Our image class
#include "../../Image/pvc_image.h"

// ----  GL / CUDA inter-operation ----

// Include GL extension header
#include <GL/glew.h>

// Include freeglut windowing
#include <GL/freeglut.h>


#include <cuda_gl_interop.h>


// this one makes it easier to exchange float for double 
// on devices of higher compute capability if more accuracy 
// is needed 
#define cuda_float float

// comment out if you have trouble with the visualziation
#define SHOW

int W;
int H;

// CUDA Block size
#define BLOCK_SIZE_X 4
#define BLOCK_SIZE_Y 4

// Macro to handle CUDA error return
#define CUDA_SAFE_CALL(c) if ( c != cudaSuccess ) { cout << "CUDA error: " << cudaGetErrorString( cudaGetLastError() )  << ", line " << __LINE__ << endl; cout.flush(); assert( false ); }


// for OpenGL interoperation
GLuint pbo_u = 0;        // pixel buffer object
GLuint pbo_v = 0;        // pixel buffer object
GLuint texture_u = 0;  // texture associated with CUDA buffer
GLuint texture_v = 0;  // texture associated with CUDA buffer

GLuint texture_left;
GLuint texture_right;

GLuint debug_program = 0;
GLuint  shader_mode  = 1;
GLfloat gammaval = 1.0;
GLfloat scale = 1.0;

bool continue_in_main_loop = true;

// --------------------------------------------------------------------------------
// ---------------------------------------- CUDA kernels  -------------------------
// --------------------------------------------------------------------------------


// sets all values in <array> to <value>
__global__ void set_value( cuda_float *array, int W, int H, cuda_float value ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ox + oy * W;

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( off >= 0 &&  off < W * H )
    array [ off ] = value;

}




// example kernel that reads input
__global__ void gauss_sep_hor( cuda_float *px, cuda_float *out, const int W, const int H, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H * 4 )
    return;
  
  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int i = -halfsize; i <= halfsize; i++ )
	  { 
	    int img_off = i * 4;

	    // use center value if kernel is outside image
	    if ( ox + i < 0 || ox + i > W - 1 )
	      img_off = 0;

	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = i * i;
	    
	    cuda_float kval = exp( - square * SIGMA_square );

	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }
  
      
}


// example kernel that reads input
__global__ void gauss_sep_ver( cuda_float *px, cuda_float *out, const int W, const int H, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H * 4 )
    return;

  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int j = -halfsize; j <= halfsize; j++ )
	  { 
	    int img_off = j * W * 4;

	    if ( oy + j < 0 || oy + j > H - 1 )
	      img_off = 0;
	    
	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = j * j;
	    
	    //	    cuda_float kval = exp( - square / ( 2.0 * SIGMA * SIGMA ) );
	    cuda_float kval = exp( - square * SIGMA_square );


	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }

}


//helper functions -- a device function is always inlined for architectures below 3.5 or so

//returns luminance of a color image
__device__ cuda_float get_luminance( cuda_float *colimg, int imgoff, int W, int H ) 
{
  // early out if memory-access is out-of-bounds 
  //if ( imgoff < 0 || imgoff > W * H * 4 )
  //  return 0.0;

  return 0.299f * colimg [ imgoff + PVC_CHANNEL_R ] + 0.587f * colimg [ imgoff + PVC_CHANNEL_G ] + 0.114f * colimg [ imgoff + PVC_CHANNEL_B ];
  //return colimg [ imgoff + PVC_CHANNEL_G ];
}

//returns "average" of a scalar field (single channel image) - required in classical formulation of Horn-Schunck
__device__ cuda_float average_classic( cuda_float *img, int off, int W, int H ) 
{

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return 0.0;

  return 1.0/12.0 * img [ off - 1 - W ] +  1.0/6.0 * img[ off - W ] + 1.0/12.0 * img [ off + 1 - W ] +
         1.0/6.0  * img [ off - 1     ] +                             1.0/6.0  * img [ off + 1     ] +
         1.0/12.0 * img [ off - 1 + W ] +  1.0/6.0 * img[ off + W ] + 1.0/12.0 * img [ off + 1 + W ];
}


// the optical flow kernel
__global__ void horn_schunck_classic_IxIyIt( cuda_float *left, cuda_float *right, int W, int H, cuda_float *Ix, cuda_float *Iy, cuda_float *It ) 
{

  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );
  int imgoff = off * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( ox >= 1 && ox < W - 1 )
    {
      // over deltaX  - choose deltaX and deltaY equal, ignore overall scale factor
      Ix [ off ] = 1.0f / 2.0f * ( get_luminance(  left, imgoff + 4, W, H ) - 
				   get_luminance(  left, imgoff - 4, W, H ) +
				   get_luminance( right, imgoff + 4, W, H ) - 
				   get_luminance( right, imgoff - 4, W, H )  );

    }
  else
    Ix [ off ] = 0.0f; // von Neumann = 0
  

  if ( oy >= 1 && oy < H - 1 )
    {
      // over deltaX  - choose deltaX and deltaY equal, ignore overall scale factor
      Iy [ off ] = 1.0f / 2.0f * ( get_luminance(  left, imgoff + 4 * W, W, H ) - 
				   get_luminance(  left, imgoff - 4 * W, W, H ) + 
				   get_luminance( right, imgoff + 4 * W, W, H ) -
				   get_luminance( right, imgoff - 4 * W, W, H ) ); 
    }
  else
    Iy [ off ] = 0.0f; // von Neumann = 0

  // ( I_{t+1} - I_{t} ) / deltaT -> again, ignore deltaT
  It [ off ] = ( get_luminance( right, imgoff, W, H ) -  get_luminance( left, imgoff, W, H ) );

}

__global__ void horn_schunck_classic_ubarvbar( cuda_float *u, cuda_float *v, int W, int H, cuda_float *ubar, cuda_float *vbar ) 
{
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  if ( ox > 0 && ox < W - 1 &&
       oy > 0 && oy < H - 1 )
    {
      ubar[ off ] = average_classic( u, off, W, H );
      vbar[ off ] = average_classic( v, off, W, H );
    }
  else
    {
      //default - corners
      ubar [ off ] = 0.0;
      vbar [ off ] = 0.0;

      // left
      if ( ox == 0 && oy != 0 && oy != H - 1 )
	{
	  ubar[ off ] = average_classic( u, off + 1, W, H );
	  vbar[ off ] = average_classic( v, off + 1, W, H );
	}

      // right
      if ( ox == W - 1 && oy != 0 && oy != H - 1 )
	{
	  ubar[ off ] = average_classic( u, off - 1, W, H );
	  vbar[ off ] = average_classic( v, off - 1, W, H );
	}


      // top
      if ( oy == 0 && ox != 0 && ox != W - 1 )
	{
	  ubar[ off ] = average_classic( u, off + W, W, H );
	  vbar[ off ] = average_classic( v, off + W, W, H );
	}

      // bottom
      if ( oy == H - 1 && ox != 0 && ox != W - 1 )
	{
	  ubar[ off ] = average_classic( u, off - W, W, H );
	  vbar[ off ] = average_classic( v, off - W, W, H );
	}

    }
}

__global__ void horn_schunck_classic_take_step( cuda_float *u, cuda_float *v, int W, int H, cuda_float *ubar, cuda_float *vbar, cuda_float *Ix, cuda_float *Iy, cuda_float *It, float ALPHA ) 
{

  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;
  
  int off    = ( ox + oy * W );
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  cuda_float vx  = Ix [ off ];
  cuda_float vy  = Iy [ off ];

  cuda_float val = ( vx * ubar[ off ]  +
		     vy * vbar[ off ]  +
		     It [ off ]  ) / ( ALPHA * ALPHA + vx * vx + vy * vy );

  u [ off ]  = ubar [ off ] - vx * val;
  
  v [ off ]  = vbar [ off ] - vy * val;
  
}




// the visualization kernel - takes the scalar fields u and v (components of optical flow)
// and turns it into a color image in the range [0,1]
__global__ void visualize( cuda_float *u, cuda_float *v, cuda_float *visresult, int W, int H )
{
  cuda_float MAXLENGTH = 5.0;
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );
  int imgoff = off * 4;
  
  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  // simple 
  visresult [ imgoff + PVC_CHANNEL_R ] = u [ off ] / MAXLENGTH + 0.5;
  visresult [ imgoff + PVC_CHANNEL_G ] = v [ off ] / MAXLENGTH + 0.5;
  visresult [ imgoff + PVC_CHANNEL_B ] = 0.0;
  visresult [ imgoff + PVC_CHANNEL_A ] = 1.0;

  // HSV encoding
  if ( 0 )
    {
      cuda_float length = sqrt( u [ off ] * u [ off ] + v [ off ] * v [ off ] );
      
      cuda_float cosa = u [ off ] / length;
      cuda_float sina = v [ off ] / length;
      
      cuda_float angle = atan2( cosa, sina );   //C
      
      cuda_float value = length / MAXLENGTH;    //H 
      
      cuda_float angle_dash = angle / (3.14159f / 3.0f );
      
      if ( 0.0f <= angle_dash && angle_dash < 1.0f )
	{
	  visresult [ imgoff + 2 ] = u [ off ] * 10.0 + 0.5;
	  visresult [ imgoff + 1 ] = v [ off ] * 10.0 + 0.5;
	  visresult [ imgoff + 0 ] = 0.0;
	  visresult [ imgoff + 3 ] = 1.0;
	  
	}
    }
} 




//returns bilinearly interpolated pixel value
__device__ cuda_float get_pixel_bilinear( cuda_float *img, cuda_float x, cuda_float y, int W, int H, int c ) 
{
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return 0.0;

  if ( x < 0 || x > W - 1 || y < 0 || y > H - 1 )
    return 0.0;

  off = ( floor( x ) + floor( y ) * W ) * 4 + c;

  cuda_float Iul, Iur, Ill, Ilr;

  Iul = img [ off ];
  Iur = img [ off + 4 ];
  Ill = img [ off + 4 * W ];
  Ilr = img [ off + 4 * W + 4 ];

  cuda_float ax, ay;
  
  ax =  x - floor( x );
  ay =  y - floor( y );

  cuda_float Iux, Ilx;

  Iux = Iul * ( 1.0f - ax ) + Iur * ax;
  Ilx = Ill * ( 1.0f - ax ) + Ilr * ax;
  
  return ( 1.0f - ay ) * Iux + ay * Ilx;
}





// the warp kernel - uses the scalar fields u and v (components of optical flow)
// to warp the from image 
__global__ void warp( cuda_float *u, cuda_float *v, cuda_float *from, cuda_float *visresult, int W, int H )
{
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off    = ( ox + oy * W );

  // early out if memory-access is out-of-bounds 
  if ( off < 0 || off > W * H )
    return;

  cuda_float uval = ox + u [ off ] * 2;
  cuda_float vval = oy + v [ off ] * 2;


  int imgoff       = off * 4;

  if ( uval > 0 && uval < W - 1 && 
       vval > 0 && vval < H - 1 )
    {
      /*
	int imglookupoff = ( ox + uval + (oy +vval) * W ) * 4; 
	
	// nearest
	visresult [ imgoff + PVC_CHANNEL_R ] = from  [ imglookupoff + PVC_CHANNEL_R ];
	visresult [ imgoff + PVC_CHANNEL_G ] = from  [ imglookupoff + PVC_CHANNEL_G ];
	visresult [ imgoff + PVC_CHANNEL_B ] = from  [ imglookupoff + PVC_CHANNEL_B ];
	visresult [ imgoff + PVC_CHANNEL_A ] = from  [ imglookupoff + PVC_CHANNEL_A ];
      */

	// bilinear
      visresult [ imgoff + PVC_CHANNEL_R ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_R  );
      visresult [ imgoff + PVC_CHANNEL_G ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_G  );
      visresult [ imgoff + PVC_CHANNEL_B ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_B  );
      visresult [ imgoff + PVC_CHANNEL_A ] = get_pixel_bilinear ( from, uval, vval, W, H, PVC_CHANNEL_A  );

    }
  else
    {
      visresult [ imgoff + PVC_CHANNEL_R ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_G ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_B ] = 0.0f;
      visresult [ imgoff + PVC_CHANNEL_A ] = 0.0f;
    }
} 



// --------------------------------------------------------------------------------
// ------------------------------------- OpenGL stuff  ----------------------------
// --------------------------------------------------------------------------------

void createPBO( GLuint* pbo, int W, int H, int channels )
{
 
  if (pbo) {
    // set up vertex data parameter
    int num_texels = W * H;
    int num_values = num_texels * channels;
    int size_tex_data = sizeof(GLfloat) * num_values;
     
    // Generate a buffer ID called a PBO (Pixel Buffer Object)
    glGenBuffers( 1, pbo );
    // Make this the current UNPACK buffer (OpenGL is state-based)
    glBindBuffer( GL_PIXEL_UNPACK_BUFFER, *pbo );
    // Allocate data for the buffer. 1-channel float image
    glBufferData( GL_PIXEL_UNPACK_BUFFER, size_tex_data, NULL, GL_DYNAMIC_COPY );
    
  }

  // this one kills CUDA kernel launches if the pixel buffer object is 
  // not handed to CUDA afterwards (my interpretation: GL takes over)
  //  cudaGLRegisterBufferObject( *pbo );


}
 
void deletePBO( GLuint* pbo )
{
  if (pbo) {
    // unregister this buffer object with CUDA
    cudaGLUnregisterBufferObject( *pbo );
     
    glBindBuffer( GL_ARRAY_BUFFER, *pbo);
    glDeleteBuffers( 1, pbo);
     
    *pbo = NULL;
  }
}
 
void createTexture( GLuint* textureID, unsigned int W, unsigned int H, unsigned int channels )
{
  // Enable Texturing
  glEnable(GL_TEXTURE_2D);
 
  // Generate a texture identifier
  glGenTextures( 1, textureID );
 
  // Make this the current texture (remember that GL is state-based)
  glBindTexture( GL_TEXTURE_2D, *textureID );
 
  // Allocate the texture memory. The last parameter is NULL since we only
  // want to allocate memory, not initialize it
  // use internal 32bit float format so we can analyze the data in a shader unaltered

  if ( channels == 1 )
    glTexImage2D( GL_TEXTURE_2D, 0, GL_R32F, W, H, 0,
		  GL_RED, GL_FLOAT, NULL);
  else
    if ( channels == 4 )
      glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA32F, W, H, 0,
		    GL_RGBA, GL_FLOAT, NULL);
    else
      cerr << "createTexture: number of channels is supposed to be 1 or 4 (current: " << channels << ")" << endl;
  
  // Must set the filter mode, GL_LINEAR enables interpolation when scaling
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  // Note: GL_TEXTURE_RECTANGLE_ARB may be used instead of
  // GL_TEXTURE_2D for improved performance if linear interpolation is
  // not desired. Replace GL_LINEAR with GL_NEAREST in the
  // glTexParameteri() call
 
}
 
void deleteTexture( GLuint* tex )
{
  glDeleteTextures( 1, tex );
  
  *tex = NULL;
}



// Load a file, e.g. a shader as a string
string loadFileAsString(const string& fileName) {
    
	ifstream ifs(fileName.c_str(), ios::in | ios::binary | ios::ate);

    ifstream::pos_type fileSize = ifs.tellg();
    ifs.seekg(0, ios::beg);

    vector<char> bytes((unsigned int)fileSize);
    ifs.read(&bytes[0], fileSize);

    return string(&bytes[0], (unsigned int)fileSize);
}

// Load an image into a texture and set the minimal state
static GLuint loadTexture(const PVC_Image* const pvc_image) {

	GLuint texture;
	glGenTextures(1, &texture);

	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, pvc_image->w(), pvc_image->h(), 0, GL_BGRA, GL_FLOAT, pvc_image->get_data());

	return texture;
}

static bool fileExists(const string& filename) {
	ifstream file(filename.c_str(), std::ios::in);
	return !file.fail();
}

static void dumpInfoLog(const string& infoLog) {
	cerr << "------------------------------------" << endl;
	cerr << "------------------------------------" << endl;
	cerr << infoLog;
	cerr << "------------------------------------" << endl;
	cerr << "------------------------------------" << endl;
}

static string getShaderInfoLog(const GLuint shader) {
	const int maxInfoLogSize = 1000;	
	GLchar infoLog[maxInfoLogSize];
	GLint infoLogSize;
	glGetShaderInfoLog(shader, maxInfoLogSize, &infoLogSize, infoLog);
	return infoLog;
}

static string getProgramInfoLog(const GLuint program) {
	const int maxInfoLogSize = 1000;	
	GLchar infoLog[maxInfoLogSize];
	GLint infoLogSize;
	glGetProgramInfoLog(program, maxInfoLogSize, &infoLogSize, infoLog);
	return infoLog;
}

static bool didCompile(const GLhandleARB shader) {
	int param;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &param);
	return GL_TRUE  == param;
}

static bool didLink(const GLhandleARB program) {
	int param;
	glGetProgramiv(program, GL_LINK_STATUS, &param);
	return GL_TRUE  == param;
}

// Load a single shader (faragment or vertex or ..) from a file
static GLhandleARB loadShader(const string& filename, const GLenum shaderType) {
	const GLhandleARB shader = glCreateShaderObjectARB(shaderType);

	const string source = loadFileAsString(filename);
	const GLcharARB* source2 = source.c_str();

	glShaderSourceARB(shader, 1, &source2, NULL);

	glCompileShader(shader);

	if(!didCompile(shader)) {
		//cout << "The shader '" << filename << "' did not compile successfully" << endl;		
		//dumpInfoLog(getShaderInfoLog(shader));
	}

	return shader;
}

static void attachShader(const string& filename, const GLuint program, const GLenum shaderType) {
	if(fileExists(filename)) {
		glAttachObjectARB(program, loadShader(filename, shaderType));
	}
}

// Load a "shader" (to be precise: a program with a fragment shader) using a naming convention & error checking
static GLuint loadProgram(const string& filename) {
    
	const GLuint program = glCreateProgramObjectARB();

	attachShader(filename + "Fragment.glsl", program, GL_FRAGMENT_SHADER);
	attachShader(filename + "Vertex.glsl", program, GL_VERTEX_SHADER);
	attachShader(filename + "Geometry.glsl", program, GL_GEOMETRY_SHADER);

	glLinkProgramARB(program);	

	if(!didLink(program)) {
		cout << "The program '" << filename << "' did not link successfully" << endl;		
		dumpInfoLog(getProgramInfoLog(program));
	}

	return program;
}

// Wrapper funtions to set uniform values more easily

static void setUniformFloat(const GLuint program, const string& name, const float x) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x};
	glUniform1fv(location, 1, values);
}

static void setUniformInt(const GLuint program, const string& name, const int x) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const int values[] = {x};
	glUniform1iv(location, 1, values);
}

static void setUniformFloat2(const GLuint program, const string& name, const float x, const float y) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x,y};
	glUniform2fv(location, 1, values);
}

static void setUniformFloat3(const GLuint program, const string& name, const float x, const float y, const float z) {
	const GLint location = glGetUniformLocation(program, name.c_str());
	const float values[] = {x,y,z};
	glUniform3fv(location, 1, values);
}




static void Key(unsigned char key, int x, int y) 
{
  switch ( key ) 
    {
    case '+': gammaval+=0.1; cerr << "Gammaval now: " << gammaval << endl;break;
    case '-': gammaval-=0.1; cerr << "Gammaval now: " << gammaval << endl;break;
    case '*': scale*=0.9; cerr << "Scale now: " << scale << endl;break;
    case '/': scale/=0.9; cerr << "Scale now: " << scale << endl;break;
    case 'r': debug_program = loadProgram("Debug");break;
    case '1': shader_mode = 1;break;
    case '2': shader_mode = 2;break;
    case '3': shader_mode = 3;break;
    case '4': shader_mode = 4;break;
    case '5': shader_mode = 5;break;
    case '6': shader_mode = 6;break;
    case '7': shader_mode = 7;break;
    case 'h' : 
      cerr << "Help" << endl << "----" << endl << endl;
      cerr << "+   - increase gamma by 0.1" << endl;
      cerr << "-   - decrease gamma by 0.1" << endl;
      cerr << "*   - scale values by  90% of the original value" << endl;
      cerr << "/   - scale values by 111% of the original value" << endl;
      cerr << "r   - reload shader code" << endl;
      cerr << "1   - u-component of flow (red=negative, green=positive, dark blue=Inf-values, light blue=NaN values" << endl;
      cerr << "2   - v-component of flow (red=negative, green=positive, dark blue=Inf-values, light blue=NaN values" << endl;
      cerr << "3   - middlebury colors" << endl;
      cerr << "4   - IPOL colors" << endl;
      cerr << "5   - left image (t)" << endl;
      cerr << "6   - right image (t+1)" << endl;
      cerr << "7   - warped right image (should match left image)" << endl;
      cerr << "h   - help" << endl;break;
    case 27:
      continue_in_main_loop = !continue_in_main_loop;
      break;
    }
}

static void Reshape(int width, int height) {  
  glViewport(0, 0, width, height);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity() ;
  glScalef(2, 2, 1);
  glTranslated(-0.5, -0.5, 0);
}


static void Display() {
	glClearColor(0.5, 0.5, 0.5, 0.5);
	glClear(GL_COLOR_BUFFER_BIT);

	//drawFullScreenQuad();
	//drawTri();

	// enable texturing and current texture
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);

	// Create a texture from the buffer
	glBindBuffer( GL_PIXEL_UNPACK_BUFFER, pbo_u );
	
	// bind texture from PBO
	glBindTexture( GL_TEXTURE_2D, texture_u );
	
	
	// Note: glTexSubImage2D will perform a format conversion if the
	// buffer is a different format from the texture. We created the
	// texture with format GL_RGBA32F. In glTexSubImage2D we specified
	// GL_BGRA and GL_FLOAT. This is a fast-path combination
 
	// Note: NULL indicates the data resides in device memory
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, W, H,
			 GL_RED, GL_FLOAT, NULL );
 

	// same for v component
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindBuffer( GL_PIXEL_UNPACK_BUFFER, pbo_v );
	glBindTexture( GL_TEXTURE_2D, texture_v );
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, W, H,
			 GL_RED, GL_FLOAT, NULL );


	// set textures for left and right images
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture( GL_TEXTURE_2D, texture_left );


	glActiveTextureARB(GL_TEXTURE3_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture( GL_TEXTURE_2D, texture_right );

	glUseProgramObjectARB( debug_program );
	
	// send the variable to our fragment program 
  	setUniformInt( debug_program, "ImageSampler2D_u", 0 );
  	setUniformInt( debug_program, "ImageSampler2D_v", 1 );
  	setUniformInt( debug_program, "ImageSampler2D_left", 2 );
  	setUniformInt( debug_program, "ImageSampler2D_right", 3 );

  	setUniformInt( debug_program, "shader_mode", shader_mode );

	setUniformFloat( debug_program, "gamma", gammaval );
	setUniformFloat( debug_program, "scale", scale );
	

 
	// Draw a single Quad with texture coordinates for each vertex.
 
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f,1.0f); glVertex3f(0.0f,0.0f,0.0f);
	glTexCoord2f(0.0f,0.0f); glVertex3f(0.0f,1.0f,0.0f);
	glTexCoord2f(1.0f,0.0f); glVertex3f(1.0f,1.0f,0.0f);
	glTexCoord2f(1.0f,1.0f); glVertex3f(1.0f,0.0f,0.0f);
	glEnd();
 
	// Don't forget to swap buffers
	glutSwapBuffers();

	// indicate the display must be redrawn
	glutPostRedisplay();
}




int init_OpenGL( int w, int h, int argc, char *argv[] )
{
  glutInitWindowSize( w,  h);
  //  glutInitWindowPosition( 200, 200);
  glutInitDisplayMode( GLUT_RGBA | GLUT_SINGLE );
  glutInit(&argc, argv);
  
  glutCreateWindow( "OpenGL Test" );
  
  const GLenum glewError = glewInit();
  if (GLEW_OK != glewError) 
    {
      cerr << "Error:" << glewGetErrorString(glewError);
    }
  
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );

  glutKeyboardFunc(Key);
  glutReshapeFunc(Reshape);
  glutDisplayFunc(Display);


  debug_program = loadProgram("Debug");

  return 0;
}




// --------------------------------------------------------------------------------
// ------------------------------------- CUDA init stuff  -------------------------
// --------------------------------------------------------------------------------





// from CUDA SDK
// Beginning of GPU Architecture definitions
inline int _ConvertSMVer2Cores(int major, int minor)
{
  // Defines for GPU Architecture types (using the SM version to determine the # of cores per SM
  typedef struct {
    int SM; // 0xMm (hexidecimal notation), M = SM Major version, and m = SM minor version
    int Cores;
  } sSMtoCores;

  sSMtoCores nGpuArchCoresPerSM[] = 
    { { 0x10,  8 },
      { 0x11,  8 },
      { 0x12,  8 },
      { 0x13,  8 },
      { 0x20, 32 },
      { 0x21, 48 },
      {   -1, -1 } 
    };

  int index = 0;
  while (nGpuArchCoresPerSM[index].SM != -1) {
    if (nGpuArchCoresPerSM[index].SM == ((major << 4) + minor) ) {
      return nGpuArchCoresPerSM[index].Cores;
    }
    index++;
  }
  printf("MapSMtoCores undefined SMversion %d.%d!\n", major, minor);
  return -1;
}
// end of GPU Architecture definitions




// this one checks if your card is suitable and drivers are working
int init_CUDA()
{
  // Check available devices
  int error = 0;
  int deviceCount;
  int gpuDeviceCount = 0;
  struct cudaDeviceProp deviceProp;
  
  cudaError_t result = cudaGetDeviceCount( &deviceCount );

  if ( result != cudaSuccess )
    {
      cout << "No CUDA devices found - Error string: " << cudaGetErrorString( result ) << endl;
      
      // machines without GPU should still support emulation
      deviceCount = 0; 
      
      error = -1;
    }
  
  for ( int device = 0; device < deviceCount; ++device ) 
    {
      CUDA_SAFE_CALL ( cudaGetDeviceProperties(&deviceProp, device) );
      
      if ( deviceProp.major != 9999) /* 9999 means emulation only */
	++gpuDeviceCount;
    }
  
  cout <<  gpuDeviceCount << " GPU CUDA device(s) found" << endl;

  // Use first device found - CUDA only
  CUDA_SAFE_CALL ( cudaSetDevice( 0 ) );

  // Use first device found - CUDA and GL
  //  CUDA_SAFE_CALL ( cudaGLSetGLDevice( 0 ) );

  CUDA_SAFE_CALL ( cudaSetDeviceFlags(  cudaDeviceMapHost ) );

  cudaGetDeviceProperties( &deviceProp, 0 );
  cout << "Using CUDA device " << deviceProp.name << endl;

  printf("[%s] has %d MP(s) x %d (Cores/MP) = %d (Cores)\n",
	 deviceProp.name, deviceProp.multiProcessorCount,
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor),
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor) * deviceProp.multiProcessorCount);

  cout << "ClockSpeed: " << deviceProp.clockRate << " Hz (?)." << endl;

  cout.flush();

#ifdef SHOW
  // OpenGL interoperation
  createPBO( &pbo_u, W, H, 1 );
  createTexture( &texture_u, W, H, 1 );

  createPBO( &pbo_v, W, H, 1 );
  createTexture( &texture_v, W, H, 1 );
#endif

  return error;
}


int cudaCheckError( char *desc )
{
  // report errors in this variable
  cudaError_t err;
  
  if ( (err = cudaGetLastError() ) != cudaSuccess ) 
    {
      cout << "Error at " << desc << ": " <<  cudaGetErrorString( err ) << endl;  
      cout.flush();
    }
  return err;
  
}






// --------------------------------------------------------------------------------
// ------------------------------------- main function ----------------------------
// --------------------------------------------------------------------------------

// Main function for example
int main( int argc, char *argv [] )
{
  
  
  // this is where our result will be stored
  PVC_Image  *result       = NULL;
  PVC_Image  *input_left   = NULL;
  PVC_Image  *input_right  = NULL;

  // a CUDA GPU-based array 
  cuda_float *gpu_left   = NULL;
  cuda_float *gpu_right  = NULL;

  cuda_float *gpu_u      = NULL;
  cuda_float *gpu_v      = NULL;

  cuda_float *gpu_ubar   = NULL;
  cuda_float *gpu_vbar   = NULL;

  cuda_float *gpu_Ix     = NULL;
  cuda_float *gpu_Iy     = NULL;
  cuda_float *gpu_It     = NULL;


  cuda_float *gpu_dum    = NULL;
  cuda_float *gpu_out    = NULL;

  
  // ------------ *NEW* begin --------------
  input_left = new PVC_Image;
  //input_left -> load("../../Data/teddy_left.ppm");
  //input_left -> load("a.ppm");
  input_left -> load("../../Data/yosemite10.ppm");

  input_right = new PVC_Image;
  //input_right -> load("../../Data/teddy_right.ppm");
  //input_right -> load("b.ppm");
  input_right -> load("../../Data/yosemite09.ppm");

  //input_left -> save( "Output/oflow_input_left.ppm" );
  //input_right -> save( "Output/oflow_input_right.ppm" );

  // number of pixels
  W = input_left -> w();
  H = input_left -> h();
  int N = W * H;

  if ( input_left -> w() != input_right -> w() ||
       input_left -> h() != input_right -> h() )
    {
      cout << "Image size do not match, exiting ..." << endl;
      cout.flush();
      exit( -1 );
    }
       

  // ------------ *NEW* end ----------------

#ifdef SHOW
  init_OpenGL( input_left -> w(), input_left -> h(), argc, argv );

  texture_left  = loadTexture( input_left );
  texture_right = loadTexture( input_right );
#endif

  if ( init_CUDA() < 0 )
    {
      fprintf( stderr,"Could not initialize CUDA, exiting ...\n" );
      exit ( -1 );
    }




  // allocate N * 4 floats on the GPU (4 color channels) - will be the input now
  CUDA_SAFE_CALL( cudaMalloc( &gpu_left, N * sizeof( cuda_float ) * 4 ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_right, N * sizeof( cuda_float ) * 4 ) );

  // flow vectors - in case of SHOW do not allocate these, they are OpenGL buffers
#ifndef SHOW
  CUDA_SAFE_CALL( cudaMalloc( &gpu_u, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_v, N * sizeof( cuda_float ) ) );
#endif

  // average flow 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_ubar, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_vbar, N * sizeof( cuda_float ) ) );

  // gradients
  CUDA_SAFE_CALL( cudaMalloc( &gpu_Ix, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_Iy, N * sizeof( cuda_float ) ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_It, N * sizeof( cuda_float ) ) );



  // ------------ *NEW* begin --------------
  // allocate output 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_dum, N * sizeof( cuda_float ) * 4 ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_out, N * sizeof( cuda_float ) * 4 ) );
  // ------------ *NEW* end   --------------

  
  // initialize our result image
  result = new PVC_Image;
  result -> resize( W, H );


  // Set block size (in number of threads)
  dim3 dimBlock( BLOCK_SIZE_X, BLOCK_SIZE_Y );
  
  // Set grid size (in number of blocks)
  dim3 dimGrid( W / dimBlock.x, H / dimBlock.y ); 

  
  // ------------ *NEW* begin --------------


  // initialize result arrays 
#ifndef SHOW  // in case of SHOW these will be OpenGL buffers, so memory will not be allocated
  set_value <<< dimGrid, dimBlock >>> ( gpu_u, W, H, 0.0f );
  cudaCheckError( "setvalue u" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_v, W, H, 0.0f );
  cudaCheckError( "setvalue v" );
#endif

  set_value <<< dimGrid, dimBlock >>> ( gpu_ubar, W, H, 0.0f );
  cudaCheckError( "setvalue ubar" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_vbar, W, H, 0.0f );
  cudaCheckError( "setvalue vbar" );

  set_value <<< dimGrid, dimBlock >>> ( gpu_Ix, W, H, 0.0f );
  cudaCheckError( "setvalue Ix" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_Iy, W, H, 0.0f );
  cudaCheckError( "setvalue Iy" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_It, W, H, 0.0f );
  cudaCheckError( "setvalue It" );




#ifdef SHOW
  // this one kills CUDA kernel launches if the pixel buffer object is 
  // not handed to CUDA afterwards (my interpretation: GL takes over)
  // this call is very slow according to documentation - call only once
  cudaGLRegisterBufferObject( pbo_u );
  cudaGLRegisterBufferObject( pbo_v );

  // hand memory of pbo's to CUDA
  cudaGLMapBufferObject((void**)&gpu_u, pbo_u );
  cudaGLMapBufferObject((void**)&gpu_v, pbo_v );

  // initialize result arrays 
  set_value <<< dimGrid, dimBlock >>> ( gpu_u, W, H, 0.0f );
  cudaCheckError( "setvalue u" );
  set_value <<< dimGrid, dimBlock >>> ( gpu_v, W, H, 0.0f );
  cudaCheckError( "setvalue v" );

  // hand memory of pbo's back to GL
  cudaGLUnmapBufferObject( pbo_u );
  cudaGLUnmapBufferObject( pbo_v );

#endif

  // ------------ *NEW* end ----------------

  
  // Perform computation
  cerr << "GPU computation: ["; cerr.flush();
  clock_t gpu_start = clock();  

#ifdef SHOW
  // map OpenGL buffer object for writing from CUDA on a single GPU
  // no data is moved (Win & Linux). When mapped to CUDA, OpenGL
  // should not use this buffer
  cudaGLMapBufferObject((void**)&gpu_u, pbo_u );
  cudaGLMapBufferObject((void**)&gpu_v, pbo_v );
#endif

#define ALPHA 1.0f
#define INNER_LOOPS 2000


  float SIGMA = 0.8;
  int HS = (int) ceil( SIGMA ) * 2 + 1;
  
  cerr << "SIGMA: " << SIGMA << " HalfSize: " << HS << endl;

  
  // Copy images to GPU (host to device)
  CUDA_SAFE_CALL( cudaMemcpy( gpu_left,  input_left -> get_data(), N * sizeof(cuda_float) * 4, cudaMemcpyHostToDevice ));
  CUDA_SAFE_CALL( cudaMemcpy( gpu_right, input_right -> get_data(), N * sizeof(cuda_float) * 4, cudaMemcpyHostToDevice ));
  
  
  // pre-blur left and right images
  gauss_sep_hor <<< dimGrid, dimBlock >>> ( gpu_left, gpu_dum, W, H, HS, SIGMA );
  cudaCheckError( "horizontal preblur left" );
  cudaThreadSynchronize();
  gauss_sep_ver <<< dimGrid, dimBlock >>> ( gpu_dum, gpu_left, W, H, HS, SIGMA );
  cudaCheckError( "vertical preblur left" );
  
  gauss_sep_hor <<< dimGrid, dimBlock >>> ( gpu_right, gpu_dum, W, H, HS, SIGMA );
  cudaCheckError( "horizontal preblur right" );
  cudaThreadSynchronize();
  gauss_sep_ver <<< dimGrid, dimBlock >>> ( gpu_dum, gpu_right, W, H, HS, SIGMA );
  cudaCheckError( "vertical preblur right" );
  
  // make sure all threads are done
  cudaThreadSynchronize();
      
  // compute derivatives
  horn_schunck_classic_IxIyIt <<< dimGrid, dimBlock >>> ( gpu_left, gpu_right, W, H, gpu_Ix, gpu_Iy, gpu_It );
  cudaCheckError( "compute gradients (before inner loop)" );

  cudaThreadSynchronize();
      
#ifdef SHOW
  cudaGLUnmapBufferObject( pbo_u );
  cudaGLUnmapBufferObject( pbo_v );
#endif


  for ( int i = 0; i < INNER_LOOPS; i++ )
    {
#ifdef SHOW
      cudaGLMapBufferObject((void**)&gpu_u, pbo_u );
      cudaGLMapBufferObject((void**)&gpu_v, pbo_v );
#endif
      
      // compute averages
      horn_schunck_classic_ubarvbar <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, W, H, gpu_ubar, gpu_vbar );
      cudaCheckError( "Horn-Schunck ubar vbar" );
      
      // compute main step 
      horn_schunck_classic_take_step <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, W, H, gpu_ubar, gpu_vbar, gpu_Ix, gpu_Iy, gpu_It, ALPHA );
      cudaCheckError( "Horn-Schunck main step" );

#ifdef SHOW
      cudaGLUnmapBufferObject( pbo_u );
      cudaGLUnmapBufferObject( pbo_v );
      
      glutMainLoopEvent();
      
      while ( continue_in_main_loop )	  
	glutMainLoopEvent();
      
      // If you want stepping for every iteration, uncomment this line
      //	  continue_in_main_loop = true;
#endif
	}
  

  // timing
  clock_t gpu_end = clock();
  
  cout << "] done." << endl;
  cout << "GPU time: " << (gpu_end - gpu_start) / double(CLOCKS_PER_SEC) << endl;  
  

#ifdef SHOW
  continue_in_main_loop = true;
  while ( continue_in_main_loop )
    glutMainLoopEvent();
#endif



#ifdef SHOW
  cudaGLMapBufferObject((void**)&gpu_u, pbo_u );
  cudaGLMapBufferObject((void**)&gpu_v, pbo_v );
#endif


  // Copy back result to CPU
  // we can read directly to our image class because 
  // we have non-padded memory (just W*H floats)

  // Save results
  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_left , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_left.ppm" );

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_right , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_right.ppm" );

  visualize <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, gpu_left, W, H );
  cudaCheckError( "visualize" );

  cudaThreadSynchronize();

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_left , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_flow.ppm" );

  // use (u,v) to warp gpu_right such that it matches gpu_left -> store result in gpu_left
  warp <<< dimGrid, dimBlock >>> ( gpu_u, gpu_v, gpu_right, gpu_out, W, H );
  cudaCheckError( "warp" );

  cudaThreadSynchronize();

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_out , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );
  result -> save( "Output/oflow_warp.ppm" );


#ifdef SHOW
  cudaGLUnmapBufferObject( pbo_u );
  cudaGLUnmapBufferObject( pbo_v );

  deletePBO( &pbo_u );
  deletePBO( &pbo_v );
#endif


  // Cleanup
  // GPU
  CUDA_SAFE_CALL( cudaFree( gpu_left ));
  CUDA_SAFE_CALL( cudaFree( gpu_right ));

#ifndef SHOW  // in case of SHOW  do not free these, they are OpenGL buffers
  CUDA_SAFE_CALL( cudaFree( gpu_u ));
  CUDA_SAFE_CALL( cudaFree( gpu_v ));
#endif

  CUDA_SAFE_CALL( cudaFree( gpu_ubar ));
  CUDA_SAFE_CALL( cudaFree( gpu_vbar ));
  CUDA_SAFE_CALL( cudaFree( gpu_Ix ));
  CUDA_SAFE_CALL( cudaFree( gpu_Iy ));
  CUDA_SAFE_CALL( cudaFree( gpu_It ));

  // ------------ *NEW* begin --------------
  CUDA_SAFE_CALL( cudaFree( gpu_dum ));
  CUDA_SAFE_CALL( cudaFree( gpu_out ));
  // ------------ *NEW* end ----------------

  // Host
  delete result;
  delete input_left;
  delete input_right;

  return 0;
}

